#ifndef STATISTICS_H
#define STATISTICS_H

#include <math.h>
#include <numeric>
#include <vector>

void angular_error(const float *U1, const float *V1,
		  const float *U2, const float *V2, int nx, int ny,
		  float &ae, float &std_ae,
		  int border = 15, float min = -99999, float max = 99999,
		  float *mask = NULL)
{

    std::vector<float> vae;
    
    for(int y = border; y < ny-border; y++)

      for(int x = border; x < nx-border; x++)
      {

	if((U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] >= min) &&
	   (U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] <= max) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] >= min) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] <= max) &&
	   (mask == NULL || mask[x + y * nx] > 0.0 )) {

	        float num  = U1[x + y * nx] * U2[x + y * nx] +
                         V1[x + y * nx] * V2[x + y * nx] + 1.0;

	        float den1 = U1[x + y * nx] * U1[x + y * nx] +
                         V1[x + y * nx] * V1[x + y * nx] + 1.0;

	        float den2 = U2[x + y * nx] * U2[x + y * nx] +
                         V2[x + y * nx] * V2[x + y * nx] + 1.0;

	        float den  = sqrt((double)den1) * sqrt((double) den2);

            if(num <= den && den > 0.00001)
	    {
		float k = acos((double)num/den);
		vae.push_back(acos((double)num/den) * 180 / 3.1415926);
	    }
        }
    }

    float N = (float) vae.size();

    if (N > 0)
    {
        //compute the average euclidean error
        float sum = std::accumulate(vae.begin(), vae.end(), 0.0);
        ae = sum / N;

        sum = 0;
        //compute the standard deviation
        for(std::vector<float>::iterator i = vae.begin(); i != vae.end(); i++)
        {
            const float d  = *i - ae;
            const float d2 = d * d;
            sum += d2;
        }
        std_ae = sqrt(sum / N);

    }
    else

        std_ae = ae = 9999.0;
}

void angular_error_image(const float *U1, const float *V1,
		  const float *U2, const float *V2, int nx, int ny,
          float *ae, int border = 15, float min = -99999,
          float max = 99999, float *mask = NULL)
{

    for(int i = 0; i < nx * ny; i++)
        ae[i] = 0;

    for(int y = border; y < ny-border; y++)

      for(int x = border; x < nx-border; x++)
      {

	if((U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] >= min) &&
	   (U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] <= max) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] >= min) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] <= max) &&
	   (mask == NULL || mask[x + y * nx] > 0.0 )) {

            float num  = U1[x + y * nx] * U2[x + y * nx] +
                         V1[x + y * nx] * V2[x + y * nx] + 1.0;

            float den1 = U1[x + y * nx] * U1[x + y * nx] +
                         V1[x + y * nx] * V1[x + y * nx] + 1.0;

            float den2 = U2[x + y * nx] * U2[x + y * nx] +
                         V2[x + y * nx] * V2[x + y * nx] + 1.0;

            float den  = sqrt((double)den1) * sqrt((double) den2);

            if(num <= den && den > 0.00001)

                ae[x + y * nx] = acos((double)num/den) * 180 / 3.1415926;
        }
    }
}


void euclidean_error(const float *U1, const float *V1,
		  const float *U2, const float *V2, int nx, int ny,
		  float &ee, float &std_ee,
		  int border = 15, float min = -99999, float max = 99999,
		  float *mask = NULL)
{
    std::vector<float> vee;

    // compute the euclidean errors in each pixel
    for(int y = border; y < ny-border; y++)

      for(int x = border; x < nx-border; x++)
      {

	if((U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] >= min) &&
	   (U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] <= max) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] >= min) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] <= max) &&
	   (mask == NULL || mask[x + y * nx] > 0.0 )) {

	        float U = U1[x + y * nx] - U2[x + y * nx];
	        float V = V1[x + y * nx] - V2[x + y * nx];

	        vee.push_back(sqrt(U * U + V * V));
        }
     }

    float N = (float) vee.size();
    
    if (N > 0)
    {
        //compute the average euclidean error
        float sum = std::accumulate(vee.begin(), vee.end(), 0.0);
        ee = sum / N;

        sum = 0;
        //compute the standard deviation
        for(std::vector<float>::iterator i = vee.begin(); i != vee.end(); i++)
        {
            const float d  = (*i - ee);
            const float d2 = (d * d);
            sum += d2;
        }

        std_ee = sqrt(sum / N);

    }
    else

        std_ee = ee = 9999.0;
}

void euclidean_error_image(const float *U1, const float *V1,
		  const float *U2, const float *V2, int nx, int ny,
          float *ee, int border = 15, float min = -99999,
          float max = 99999, float *mask = NULL)
{
    for(int i = 0; i < nx * ny; i++)
        ee[i] = 0;

    // compute the euclidean errors in each pixel
    for(int y = border; y < ny-border; y++)

      for(int x = border; x < nx-border; x++)
      {

	if((U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] >= min) &&
	   (U1[x + y * nx] * U1[x + y * nx] + V1[x + y * nx] * V1[x + y * nx] <= max) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] >= min) &&
	   (U2[x + y * nx] * U2[x + y * nx] + V2[x + y * nx] * V2[x + y * nx] <= max) &&
	   (mask == NULL || mask[x + y * nx] > 0.0 )) {

		float U = U1[x + y * nx] - U2[x + y * nx];
		float V = V1[x + y * nx] - V2[x + y * nx];

		ee[x + y * nx] = sqrt(U * U + V * V);
        }
    }
}

#endif
